#ifndef __ERROR_H
#define __ERROR_H

#include "keywords.h"
#include "debug.h"

/************************
 * TYPES AND STRUCTURES *
 ************************/

/* ERROR-HANDLING FEATURES */

public void Abort(char *text,char *desc);
/* Print a fatal-error message on stderr and abort program execution
   Parameters
     text: a short string identifying the error
     desc: a more thorough explanation of the error */

#define FatalError(t,d) Abort((t),(d))

typedef unsigned short ERROR;
#define ERR_C(c) c ## U
#define ERR_OK   ERR_C(0)

implem ERROR last_err; /* Most recently recorded error */

#define LastErr() VALUE(last_err)
/* Return: the most recently recorded error */

#define IsErrOk() (LastErr() == ERR_OK)
/* Return: TRUE if everything is ok */

public void RecordErr(ERROR err); 
/* Record 'err' as the most recent error if no error
   has been recorded yet. Otherwise it does nothing. */ 

#define EraseErr() (last_err = ERR_OK) 
/* Erase the currently recorded error */

#define RetErr(err) return (RecordErr(err),FALSE)
/* Record error 'err' and return with FALSE from the function */

#define RetOk() return (RecordErr(ERR_OK),TRUE)
/* Record ERR_OK and return with TRUE from the function */

#endif /* __ERROR_H */


