#include "fifo.h"
                    
/******************
 * IMPLEMENTATION *
 ******************/

/* CONSTRUCTORS */

FIFO NewFifo(CSIZE capacity)
{                                                 
  FIFO p;
  
  if (capacity < CSIZE_C(1)) 
    capacity = CSIZE_C(1);

  p = (FIFO) ALLOCMEM(sizeof(struct fifo_s) + capacity * sizeof(ITEM));
  if (p != NULL) {
    FIFO_ITEM(p) = (ITEM *) (p+1);
    FIFO_CAPA(p) = capacity;
    FifoReset(p);
  }
  return p;                                       
}

/* DESTRUCTORS */

void DelFifo(FIFO fifo)
{
  FREEMEM(fifo);
}

/* MODIFIERS */

void FifoReset(FIFO fifo)
{
  CHECK(IS_FIFO(fifo));

  FIFO_SIZE(fifo) = CSIZE_C(0);
  FIFO_HEAD(fifo) = CINDX_C(0);
  FIFO_TAIL(fifo) = CINDX_OUT;
  
  CHECK(FifoIsEmpty(fifo));
}

void FifoPutLast(FIFO fifo,ITEM item)
{
  CHECK(! FifoIsFull(fifo));

  FIFO_TAIL(fifo) = (FIFO_TAIL(fifo) + CINDX_C(1)) % FIFO_CAPA(fifo);
  FIFO_ITEM(fifo)[FIFO_TAIL(fifo)] = item;
  FIFO_SIZE(fifo)++;
}

ITEM FifoRemFirst(FIFO fifo)
{
  ITEM item;

  CHECK(! FifoIsEmpty(fifo));

  item = FIFO_ITEM(fifo)[FIFO_HEAD(fifo)];
  FIFO_HEAD(fifo) = (FIFO_HEAD(fifo) + CINDX_C(1)) % FIFO_CAPA(fifo);
  FIFO_SIZE(fifo)--;
  return item;
}
