#ifndef __INDEX_H
#define __INDEX_H

#include <stdint.h>
#include <limits.h>
#include <inttypes.h>


/* GENERIC ELEMENT FOR COLLECTIONS */

typedef void *ITEM;

#define ITEM_NULL ((ITEM) NULL)

#define IS_ITEM(i) ((i) != ITEM_NULL)


/* SIZE TYPES FOR COLLECTIONS (ARRAYS, LISTS, FIFO, ...) */

typedef uintptr_t CSIZE;

#define CSIZE_MIN  UINTPTR_MIN
#define CSIZE_MAX  UINTPTR_MAX
#ifdef UINPTR_C
#define CSIZE_C(c) UINTPTR_C(c)
#else
#if __WORDSIZE == 64
#define CSIZE_C(c) c ## UL
#else
#define CSIZE_C(c) c
#endif
#endif

#define PRIdCSIZE PRIdPTR /* Signed decimal notation */
#define PRIiCSIZE PRIiPTR /* Signed decimal notation */
#define PRIuCSIZE PRIuPTR /* Unsigned decimal notation */
#define PRIoCSIZE PRIoPTR /* Octal notation */
#define PRIxCSIZE PRIxPTR /* Lowercase hexadecimal notation */
#define PRIXCSIZE PRIXPTR /* Uppercase hexadecimal notation */

#define SCNdCSIZE SCNdPTR
#define SCNiCSIZE SCNiPTR
#define SCNuCSIZE SCNuPTR
#define SCNoCSIZE SCNoPTR
#define SCNxCSIZE SCNxPTR
#define SCNXCSIZE SCNXPTR


/* INDEX TYPES FOR COLLECTIONS (ARRAYS, LISTS, FIFO, ...) */

typedef CSIZE CINDX;

#define CINDX_MIN  CSIZE_MIN
#define CINDX_MAX  (CSIZE_MAX - 1)
#define CINDX_OUT  CSIZE_MAX  /* Invalid index */
#define CINDX_C(c) CSIZE_C(c)

#define IS_CINDX(i) ((i) < CINDX_OUT)

#define RANGE(lo,i,hi) (((lo) <= (i)) && ((i) <= (hi)))
/* Return: TRUE if 'lo <= 'i' <= 'hi' */

#define PRIdCINDX PRIdCSIZE  /* Signed decimal notation */
#define PRIiCINDX PRIiCSIZE  /* Signed decimal notation */
#define PRIuCINDX PRIuCSIZE  /* Unsigned decimal notation */
#define PRIoCINDX PRIoCSIZE  /* Octal notation */
#define PRIxCINDX PRIxCSIZE  /* Lowercase hexadecimal notation */
#define PRIXCINDX PRIXCSIZE  /* Uppercase hexadecimal notation */

#define SCNdCINDX SCNdCSIZE 
#define SCNiCINDX SCNiCSIZE 
#define SCNuCINDX SCNuCSIZE 
#define SCNoCINDX SCNoCSIZE 
#define SCNxCINDX SCNxCSIZE 
#define SCNXCINDX SCNXCSIZE
 
#endif /* __INDEX_H */


