#ifndef __INTEGER_H
#define __INTEGER_H

#include <stdint.h>
#include <limits.h>
#include <inttypes.h>

#if __WORDSIZE == 32 

/* Signed. */
typedef int32_t bint_t; /* basic int */
typedef int16_t sint_t; /* small int, half the size of an int */
typedef int64_t lint_t; /* large int, twice the size of an int */
/* Unsigned. */
typedef uint32_t ubint_t; 
typedef uint16_t usint_t;
typedef uint64_t ulint_t;

/* Minimum of signed integers. */
#define SINT_MIN INT16_MIN
#define BINT_MIN INT32_MIN
#define LINT_MIN INT64_MIN 
/* Minimum of signed integers. */
#define USINT_MIN UINT16_MIN
#define UBINT_MIN UINT32_MIN
#define ULINT_MIN UINT64_MIN 
/* Maximum of unsigned integers. */
#define SINT_MAX INT16_MAX
#define BINT_MAX INT32_MAX
#define LINT_MAX INT64_MAX
/* Maximum of unsigned integers. */
#define USINT_MAX UINT16_MAX
#define UBINT_MAX UINT32_MAX
#define ULINT_MAX UINT64_MAX
/* Constants for signed integers. */
#define SINT_C(c) INT16_C(c)
#define BINT_C(c) INT32_C(c)
#define LINT_C(c) INT64_C(c)
/* Constants for unsigned integers. */
#define USINT_C(c) UINT16_C(c)
#define UBINT_C(c) UINT32_C(c)
#define ULINT_C(c) UINT64_C(c)

/* Macros for printing format specifiers.  */

/* Small integers. */
#define PRIdSI PRId16 /* Signed decimal notation */
#define PRIiSI PRIi16 /* Signed decimal notation */
#define PRIuSI PRIu16 /* Unsigned decimal notation */
#define PRIoSI PRIo16 /* Octal notation */
#define PRIxSI PRIx16 /* Lowercase hexadecimal notation */
#define PRIXSI PRIX16 /* Uppercase hexadecimal notation */
/* Basic integers. */
#define PRIdBI PRId32 /* Signed decimal notation */
#define PRIiBI PRIi32 /* Signed decimal notation */
#define PRIuBI PRIu32 /* Unsigned decimal notation */
#define PRIoBI PRIo32 /* Octal notation */
#define PRIxBI PRIx32 /* Lowercase hexadecimal notation */
#define PRIXBI PRIX32 /* Uppercase hexadecimal notation */
/* Large integers. */
#define PRIdLI PRId64 
#define PRIiLI PRIi64
#define PRIuLI PRIu64
#define PRIoLI PRIo64
#define PRIxLI PRIx64
#define PRIXLI PRIX64

/* Macros for scanning format specifiers.  */

/* Small integers. */
#define SCNdSI SCNd16
#define SCNiSI SCNi16
#define SCNuSI SCNu16
#define SCNoSI SCNo16
#define SCNxSI SCNx16
#define SCNXSI SCNX16
/* Basic integers. */
#define SCNdBI SCNd32
#define SCNiBI SCNi32
#define SCNuBI SCNu32
#define SCNoBI SCNo32
#define SCNxBI SCNx32
#define SCNXBI SCNX32
/* Large integers. */
#define SCNdLI SCNd64
#define SCNiLI SCNi64
#define SCNuLI SCNu64
#define SCNoLI SCNo64
#define SCNxLI SCNx64
#define SCNXLI SCNX64


#else /* __WORDSIZE == 64 */

/* Signed. */
typedef int64_t bint_t; /* basic int */
typedef int32_t sint_t; /* small int, half the size of a basic integer */
typedef int64_t lint_t; /* large int, twice the size of a basic integer */
/* Unsigned. */
typedef uint32_t usint_t;
typedef uint64_t ubint_t; 
typedef uint64_t ulint_t;

/* Minimum of signed integers. */
#define SINT_MIN INT32_MIN
#define BINT_MIN INT64_MIN
#define LINT_MIN INT64_MIN 
/* Minimum of unsigned integers. */
#define USINT_MIN UINT32_MIN
#define UBINT_MIN UINT64_MIN
#define ULINT_MIN UINT64_MIN 
/* Maximum of signed integers. */
#define SINT_MAX INT32_MAX
#define BINT_MAX INT64_MAX
#define LINT_MAX INT64_MAX
/* Maximum of unsigned integers. */
#define USINT_MAX UINT32_MAX
#define UBINT_MAX UINT64_MAX
#define ULINT_MAX UINT64_MAX
/* Constants for signed integers. */
#define SINT_C(c) INT32_C(c)
#define BINT_C(c) INT64_C(c)
#define LINT_C(c) INT64_C(c)
/* Constants for unsigned integers. */
#define USINT_C(c) UINT32_C(c)
#define UBINT_C(c) UINT64_C(c)
#define ULINT_C(c) UINT64_C(c)

/* Macros for printing format specifiers.  */

/* Small integers. */
#define PRIdSI PRId32 /* Signed decimal notation */
#define PRIiSI PRIi32 /* Signed decimal notation */
#define PRIuSI PRIu32 /* Unsigned decimal notation */
#define PRIoSI PRIo32 /* Octal notation */
#define PRIxSI PRIx32 /* Lowercase hexadecimal notation */
#define PRIXSI PRIX32 /* Uppercase hexadecimal notation */
/* Basic integers. */
#define PRIdBI PRId64 /* Signed decimal notation */
#define PRIiBI PRIi64 /* Signed decimal notation */
#define PRIuBI PRIu64 /* Unsigned decimal notation */
#define PRIoBI PRIo64 /* Octal notation */
#define PRIxBI PRIx64 /* Lowercase hexadecimal notation */
#define PRIXBI PRIX64 /* Uppercase hexadecimal notation */
/* Large integers. */
#define PRIdLI PRId64 
#define PRIiLI PRIi64
#define PRIuLI PRIu64
#define PRIoLI PRIo64
#define PRIxLI PRIx64
#define PRIXLI PRIX64

/* Macros for scanning format specifiers.  */

/* Small integers. */
#define SCNdSI SCNd32
#define SCNiSI SCNi32
#define SCNuSI SCNu32
#define SCNoSI SCNo32
#define SCNxSI SCNx32
#define SCNXSI SCNX32
/* Basic integers. */
#define SCNdBI SCNd64
#define SCNiBI SCNi64
#define SCNuBI SCNu64
#define SCNoBI SCNo64
#define SCNxBI SCNx64
#define SCNXBI SCNX64
/* Large integers. */
#define SCNdLI SCNd64
#define SCNiLI SCNi64
#define SCNuLI SCNu64
#define SCNoLI SCNo64
#define SCNxLI SCNx64
#define SCNXLI SCNX64

#endif 

#endif /* __INTEGER_H */
