#ifndef __KEYWORDS_H
#define __KEYWORDS_H

/* SPECIAL KEYWORDS:
     those keywords are used to add semantical meaning to the declarations
     made in the header of a module:
     - 'feature' qualifies properties of a module (variable or function)
     - 'deferred' qualifies features that are declared in the module but 
       not implemented in it. They should be later implemented by modules 
       which include the header in which they are declared.
     - 'public' qualifies features that are part of the interface.
        They can be freely used by modules which include (inherit) them.
     - 'implem' qualifies features that belong to the implementation of a 
        module. They must not be used outside of the module itself.
   CONVENTIONS: 
     - the keyword 'feature' is optional 
     - unqualified features are considered public
*/

#define feature /* feature */
#define deferred /* deferred */
#define public /* public */ extern
#define implem /* private */ extern
#define private /*private */ extern

#endif /* __KEYWORDS_H */ 
