#ifndef __LIST_H
#define __LIST_H

#include "platform.h"
#include "index.h"

/************************
 * TYPES AND STRUCTURES *
 ************************/

typedef struct lnode_s lnode_t, *LNODE;
struct lnode_s
{
  LNODE nd_succ; /* successor */
  LNODE nd_pred; /* predecessor */
  ITEM  nd_item; /* content */
};

#define LNODE_SUCC(n) ((n)->nd_succ)
#define LNODE_PRED(n) ((n)->nd_pred)
#define LNODE_ITEM(n) ((n)->nd_item)

typedef struct list_s list_t, *LIST;
struct list_s
{
  LNODE li_head; /* first node */
  LNODE li_pred; /* fake predecessor (must always be NULL) */
  LNODE li_tail; /* last node */
  CSIZE li_size; /* number of items in the list */
  CINDX li_fidx; /* fast access */
  LNODE li_fnod; /* fast access */
};

#define LIST_HEAD(l) ((l)->li_head)
#define LIST_PRED(l) ((l)->li_pred)
#define LIST_TAIL(l) ((l)->li_tail)
#define LIST_SIZE(l) ((l)->li_size)
#define LIST_FAST_INDEX(l) ((l)->li_fidx)
#define LIST_FAST_LNODE(l) ((l)->li_fnod)
#define LIST_HEADER(l) ((LNODE) &(LIST_HEAD(l)))
#define LIST_TAILER(l) ((LNODE) &(LIST_PRED(l)))

/************
 * FEATURES *
 ************/

/* ASSERTIONS */

#define IS_LIST(list) (list != NULL)

#define IS_LIDX(list,idx) (IS_LIST(list) && ((idx) < LIST_SIZE(list)))

/* CONSTRUCTORS AND DESTRUCTORS */

public LIST NewList(void);

public void DelList(LIST list);

/* ACCESSORS */

#define ListSize(list) GETSAFE(IS_LIST(list),LIST_SIZE(list))

#define ListIsEmpty(list) (ListSize(list) <= CSIZE_C(0))

#define ListLastIndex(list) (ListSize(list) - CINDX_C(1))

public ITEM ListItem(LIST list,CINDX idx);

#define ListFirstItem(list) ListItem((list),CINDX_C(0))

#define ListLastItem(list) ListItem((list),ListLastIndex(list))

public CINDX ListFind(LIST list,ITEM item);

#define ListHas(list,item) (ListFind((list),(item)) != CINDX_OUT)

/* MODIFIERS */

public BOOL ListPut(LIST list,ITEM item,CINDX idx);

#define ListPutFirst(list,item) ListPut((list),(item),CINDX_C(0))

#define ListPutLast(list,item) ListPut((list),(item),ListSize(list))

public ITEM ListSet(LIST list,ITEM item,CINDX idx);

#define ListSetFirst(list,item) ListSet((list),(item),CINDX_C(0))

#define ListSetLast(list,item) ListSet((list),(item),ListLastIndex(list))

public void ListRemove(LIST list,CINDX idx);

#define ListRemFirst(list) ListRemove((list),CINDX_C(0))

#define ListRemLast(list) ListRemove((list),ListLastIndex(list))

public void ListRemItem(LIST list,ITEM item);

public void ListSwap(LIST list,CINDX idx1,CINDX idx2);

public void ListShiftL(LIST list);

public void ListShiftR(LIST list);

public void ListFlush(LIST list);

/* DISPLAY */

typedef void (*PFUNC)(ITEM);

public void ListPrint(LIST list,PFUNC print_item);

#endif



