#ifndef __PLATFORM_H
#define __PLATFORM_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "keywords.h"
#include "boolean.h"
#include "integer.h"
#include "error.h"


/* DISPLAY FEATURES */

#define IsTTY() isatty(STDIN_FILENO)
/* Return: TRUE if the standard input is a terminal */

#define NewLine() printf("\n")
/* Write a newline on stdout */

#define PrintN(s,n) { unsigned int i; for (i = 0u; i < (n); i++) printf(s); }
/* Write the string 's' 'n' times on stdout */ 

#define PrintSpace(n) PrintN(" ",(n))
/* Write 'n' whitespaces on stdout */


/* MEMORY-HANDLING FEATURES */

#define ALLOCMEM(size) calloc((size),1)
/* Allocate 'size' bytes of memory
   Return: - a void pointer to the newly allocated memory 
           - NULL if not enough memory available 
*/

#define FILLMEM(mem,lgt,car) memset((mem),(car),(lgt)) 
/* Fill 'lgt' bytes of memory, starting at address 'mem',
   with the character 'car' */

#define ZEROMEM(mem,lgt) FILLMEM((mem),(lgt),0)
/* Fill 'lgt' bytes of memory starting at address 'mem' with zeros */

#define FREEMEM(mem) (free(mem), mem = NULL)
/* Free previously allocated memory 'mem' */


/* PATHNAME-HANDLING FEATURES */

#define MAXPATHNAME 511 /* excluding the terminating '\0' character */

size_t StrCpyN(char *dest,char *src,size_t n);

size_t StrCatN(char *dest,char *src,size_t n);

size_t StrFmtPath(char *path);


#endif
