/***************************************************************************
 * ENGINE.C: implements the skeleton of the proof-search automaton.        *
 ***************************************************************************/

#include <unistd.h>
#include <sys/time.h>
#include <sys/resource.h>
#include "engine.h"

BOOL abort_search;

void Search(unsigned int n)
{
  struct rusage usage;
  milsec_t start_time;
  milsec_t stop_time;

  if (n < 1U) n = 1U; /* Check parameter: one search at least */

  REQUIRE(IS_FOREST()); /* Check that there is a forest */
  REQUIRE(IS_PSTACK()); /* Check that there is proof-stack */
  REQUIRE(IS_ENGINE()); /* Check that there is an engine */

  method = &tmet[search_typmet]; ENSURE(IS_METHOD()); /* Set search method */
  forimp = &timp[search_typimp]; ENSURE(IS_IMPLEM()); /* Set search implem */

  getrusage(RUSAGE_SELF,&usage);
  start_time = ((milsec_t) usage.ru_utime.tv_sec) * MILSEC_C(1000) 
            + ((milsec_t) usage.ru_utime.tv_usec) / MILSEC_C(1000);
  
  abort_search = FALSE;
  search_iter = 0U;

 START:
  search_iter++; /* Count one entrance in the search loop */
  search_typres = TYPRES_OUT; /* Reset search-result */
  ResetPs(); /* Reset the proof-stack before starting a new iteration */
  ForestInit(); /* Initialiwe forest implementation */
  MethodInit(); /* Initialize the proof-search method */
  EngineStart(); /* Start the engine */
  
 TESTAX:
  if (abort_search) goto ABORT;
  else if (! SeqIsAxiom()) goto SELECT; 
  else goto EXPSUCC;
  
 SELECT:
  EngineSelect();
  if (select_fnidx != FINDX_OUT) goto EXPAND;
  else goto EXPFAIL;
  
 EXPAND:
  /* There must be an index to expand and a corresponding rule to apply */
  REQUIRE(IS_FNODE(select_fnidx));
  REQUIRE(RuleIsPrem(select_rule,select_prem));
  EngineExpand();
  /* Apply the selected rule to the selected index */
  PushAndApply(select_rule,select_prem,select_fnidx);
  goto TESTAX;
  
 EXPSUCC: 
  /* We have reached an axiomatic forest */
  EngineExpandSucc();
  
 REVSUCC:
  if (PsSize() <= 1) goto SUCCESS;
  /* Reverse top premiss and get back premiss results */
  ReverseAndPop();
  /* Let the engine decide whether to restart or not */
  EnginePremissSucc();
  if (select_fnidx == FINDX_OUT) goto REVSUCC; 
  else goto EXPAND;
  
 EXPFAIL:
  /* We have reached an irreducible non-axiomatic forest */
  EngineExpandFail();
  
 REVFAIL:
  if (PsSize() <= 1) goto FAILURE;
  /* Reverse top premiss and get back premiss results */
  ReverseAndPop();
  /* Let the engine decide whether to restart or not */
  EnginePremissFail();
  if (select_fnidx == FINDX_OUT) goto REVFAIL; 
  else goto EXPAND;
  
 SUCCESS:
  search_typres = SUCCESS;  
  EngineSuccess(); 
  ENSURE(SearchResult() == SUCCESS);
  if (search_iter < n) goto START;
  else goto STOP;
  
 FAILURE:
  search_typres = FAILURE;
  EngineFailure(); 
  ENSURE(SearchResult() == FAILURE);
  if (search_iter < n) goto START;
  else goto STOP;

 ABORT:
  while (PsSize() > 1) ReverseAndPop();
  search_typres = TYPRES_OUT;
  abort_search = FALSE;

 STOP:
  getrusage(RUSAGE_SELF,&usage);
  stop_time = ((milsec_t) usage.ru_utime.tv_sec) * MILSEC_C(1000) 
            + ((milsec_t) usage.ru_utime.tv_usec) / MILSEC_C(1000);
  search_time = stop_time - start_time;

  ENSURE(IS_TYPMET(SearchMethod()));
  ENSURE(IS_TYPIMP(SearchImplem()));
  return;
}

