#ifndef __ENGINE_H
#define __ENGINE_H

#include "implems.h"
#include "methods.h"

/************************
 * TYPES AND STRUCTURES *
 ************************/

typedef struct engine_s engine_t, *ENGINE;
struct engine_s
{
  void (*Start)(void);
  void (*Select)(void);
  void (*Expand)(void);
  void (*ExpandSucc)(void);
  void (*ExpandFail)(void);  
  void (*PremissSucc)(void);
  void (*PremissFail)(void);
  void (*Success)(void);
  void (*Failure)(void);
};

/* HELPERS */

#define DEF_ENGINE(engname) {\
 engname ## Start,\
 engname ## Select,\
 engname ## Expand,\
 engname ## ExpandSucc,\
 engname ## ExpandFail,\
 engname ## PremissSucc,\
 engname ## PremissFail,\
 engname ## Success,\
 engname ## Failure\
}

/* ASSERTIONS */

implem deferred ENGINE engine; /* Current proof-search engine */

#define IS_ENGINE() ( engine && engine->Start \
 && engine->Select && engine->Expand\
 && engine->ExpandSucc && engine->ExpandFail\
 && engine->PremissSucc && engine->PremissFail\
 && engine->Success && engine->Failure )

/* MODIFIERS */

#define EngineStart()       engine->Start()
#define EngineSelect()      engine->Select()
#define EngineExpand()      engine->Expand()
#define EngineExpandSucc()  engine->ExpandSucc()
#define EngineExpandFail()  engine->ExpandFail()
#define EnginePremissSucc() engine->PremissSucc()
#define EnginePremissFail() engine->PremissFail()
#define EngineSuccess()     engine->Success()
#define EngineFailure()     engine->Failure()

implem void Search(unsigned int n);
/* Perform a proof-search to decide the current forest and 
   repeat it 'n' times (or only once if 'n' is lower than 1) */

/* Search parameters */

implem typmet_t search_typmet;

#define SearchMethod() VALUE(search_typmet)
/* Return: the proof-search method used for the result */

implem typimp_t search_typimp;

#define SearchImplem() VALUE(search_typimp)
/* Return: the forest implementation used for the result */


/* STRIP SEARCH-RESULT TYPES */

typedef enum { FAILURE, SUCCESS } typres_t;
#define TYPRES_MIN FAILURE
#define TYPRES_MAX SUCCESS
#define TYPRES_OUT (TYPRES_MAX + 1)
#define IS_TYPRES(r) ((r) <= TYPRES_MAX)

implem typres_t search_typres;

#define SearchResult() VALUE(search_typres)
/* Return: the value of result (SUCCESS,FAILURE) */

implem char *typres_t_str[];

#define SearchResultStr() GETSAFE(IS_TYPRES(SearchResult()),\
				  typres_t_str[SearchResult()])
/* Return: the name of the result value in clear */

implem unsigned int  search_iter;

#define SearchLoops() VALUE(search_iter)
/* Return: the number of iteration made for the result */

typedef unsigned long int milsec_t; /* Time in 10e-3 seconds */
#define MILSEC_C(c) c ## UL
#define MILSEC_MIN  MILSEC_C(0)
#define MILSEC_MAX  MILSEC_C(-1)
#define PRIMILSEC   "lu" /* Macro for printing a time in 10e-3 seconds */ 
#define SCNMILSEC   "lu" /* Macro for scanning a time in 10e-3 seconds */

implem milsec_t search_time;

#define SearchTime() VALUE(search_time)
/* Return: how long it took in milliseconds to decide the result */

#endif




