#ifndef __IMPLEM_H
#define __IMPLEM_H

#include "forest.h"

/************************
 * TYPES AND STRUCTURES *
 ************************/

typedef struct implem_s implem_t, *IMPLEM;
struct implem_s
{
  void (*Init)(void);
  void (*ExtendL)(FINDX l,FINDX s);
  void (*NewRootL)(FINDX l);
  void (*RemoveL)(FINDX l);
  void (*ReplaceL)(FINDX l,FINDX new);
  void (*SplitLH)(FINDX l,FINDX lft,FINDX rgt);
  void (*UnsplitLH)(FINDX l,FINDX lft,FINDX rgt);
  void (*SplitLV)(FINDX l,FINDX up,FINDX dw);
  void (*UnsplitLV)(FINDX l,FINDX up,FINDX dw);
  void (*AttachST)(FINDX l,FINDX fs,FINDX lml,FINDX rml);
  void (*DetachST)(FINDX l,FINDX *fs,FINDX *lml,FINDX *rml);
};

/* HELPERS */

#define DEF_IMPLEM(impname) {\
 impname ## Init,\
 impname ## ExtendL,\
 impname ## NewRootL,\
 impname ## RemoveL,\
 impname ## ReplaceL,\
 impname ## SplitLH,\
 impname ## UnsplitLH,\
 impname ## SplitLV,\
 impname ## UnsplitLV,\
 impname ## AttachST,\
 impname ## DetachST\
}

#define LinkL(l,r) (((l==FINDX_OUT) ? (N_FIRST_LEAF=(r)):(N_RL(l)=(r))),\
		    ((r==FINDX_OUT) ? (N_LAST_LEAF=(l)):(N_LL(r)=(l))))

implem deferred IMPLEM forimp; /* Current forest implementation */
implem deferred ubint_t forimp_base;
implem deferred ubint_t forimp_ovhd;

#define INIT_IMPLEM() (forimp_base = forimp_ovhd = UBINT_C(0))

/* ASSERTIONS */

#define IS_IMPLEM() ( forimp && forimp->Init \
                      && forimp->ExtendL\
		      && forimp->NewRootL\
		      && forimp->RemoveL\
		      && forimp->ReplaceL\
		      && forimp->SplitLH\
		      && forimp->UnsplitLH\
		      && forimp->SplitLV\
		      && forimp->UnsplitLV\
		      && forimp->AttachST\
		      && forimp->DetachST )


/* MODIFIERS */

#define ForestInit() (INIT_IMPLEM(), forimp->Init())

#define ForestExtendL(l,s) forimp->ExtendL((l),(s))

#define ForestNewRootL(l) forimp->NewRootL(l)

#define ForestRemoveL(l) forimp->RemoveL(l)

#define ForestReplaceL(l,n) forimp->ReplaceL((l),(n))

#define ForestSplitLH(l,ll,rl) forimp->SplitLH((l),(ll),(rl))

#define ForestUnsplitLH(l,ll,rl) forimp->UnsplitLH((l),(ll),(rl))

#define ForestSplitLV(l,u,d) forimp->SplitLV((l),(u),(d))

#define ForestUnsplitLV(l,u,d) forimp->UnsplitLV((l),(u),(d))

#define ForestAttachST(l,s,lm,rm) forimp->AttachST((l),(s),(lm),(rm))

#define ForestDetachST(l,s,lm,rm) forimp->DetachST((l),(s),(lm),(rm))

/* ACCESSORS */

#define ForestBase() VALUE(forimp_base)

#define ForestOvhd() VALUE(forimp_ovhd)

#define ForestCost() (ForestBase() + ForestOvhd())

#endif
