/* This module implements the firs-leaf strategy */

#include "metfleaf.h"

/******************
 * IMPLEMENTATION *
 ******************/

/* FUNCTIONS */

void FindNextRule(FINDX lo,FINDX hi)
/* Find the first expandable index, scanning left to right from index 'lo' 
   toward index 'hi'. If no index can be expanded in the specified range,
   the function sets the global variable 'select_fnidx' to FINDX_OUT.
   Otherwise, the selected triple (index,rule,premiss) is returned through
   the global variables 'select_fnidx', 'select_rule', 'select_prem'.
   Remark
     Specifying a range in which 'lo' is higher than 'hi' corresponds to 
     [lo,last] U [first,hi], i.e., to a circular scanning of the range.
*/
{
  RULE rulecur;
  FINDX indexcur;

  REQUIRE(IS_FNODE(lo));   
  REQUIRE(IS_FNODE(hi));

  indexcur = lo;
  method_base++; /* one direct look at index "lo" */
  while (TRUE)
    {
      rulecur = FindRule(indexcur);
      if (rulecur != RULE_OUT) {
	select_fnidx = indexcur;
	select_rule = rulecur;
	select_prem = PREM_FST;
	return;
      }
      if (indexcur == hi) {
	select_fnidx = FINDX_OUT;
	return;
      }
      indexcur = SeqNextIdx(indexcur);
      method_ovhd++; /* one overhead look at next index */
    }
}

void FLEAFInit(void) 
{ 
}

void FLEAFSelect(void)
{
  if (focus_fnidx != FINDX_OUT)
    FindNextRule(focus_fnidx,SeqPrevIdx(focus_fnidx));
  else
    FindNextRule(SeqFirstIdx(),SeqLastIdx());
}

void FLEAFRetry(void)
{
  FINDX next;

  next = SeqNextIdx(RevIdx());
  if ((next == RevIdx()) || (next == PE_START(Top()))) 
    /* Either the next index is once again the reversed index which failed
       (only one index in the context), or the next index is the first index 
       we tried for the current sequent. In each case, we have already tried 
       all the indexes of the current sequent and we cannot try anything else.
    */
    select_fnidx = FINDX_OUT;
  else
    /* There may be another index to try */
    FindNextRule(next,SeqPrevIdx(PE_START(Top())));
}

BOOL FLEAFOkCM(void)
{
  return FALSE; /* In some cases, it is possible to build a counter-model,
		   but not in general, so, return FALSE. */
}


