#ifndef __METHOD_H
#define __METHOD_H

#include "pstack.h"

/************************
 * TYPES AND STRUCTURES *
 ************************/

typedef struct method_s method_t, *METHOD;
struct method_s
{
  void (*Init)(void); 
  /* Initialize the method before starting a proof-search */
  void (*Select)(void);
  /* Select an index to expand in the current forest. 
     If there is no such index, the function puts the value FINDX_OUT 
     into the global variable 'select_fnidx'. Otherwise, the selected 
     triple (index,rule,premiss) is returned through the global variables 
     'select_fnidx', 'select_rule', 'select_prem'. */
  void (*Retry)(void);
  /* Select and index to continue the proof-search after the currently
     reversed non-invertible rule failed.
     If there is no such index, the function puts the value FINDX_OUT 
     into the global variable 'select_fnidx'. Otherwise, the selected 
     triple (index,rule,premiss) is returned through the global variables 
     'select_fnidx', 'select_rule', 'select_prem'. */
  BOOL (*OkCM)(void); 
  /* Return: TRUE if the method allow counter-model construction */
};


/* HELPERS */

#define DEF_METHOD(metname) {\
 metname ## Init,\
 metname ## Select,\
 metname ## Retry,\
 metname ## OkCM\
}   

implem deferred METHOD  method; /* Current proof-search method */
implem deferred ubint_t method_base; /* Cost of the method in the ideal case */
implem deferred ubint_t method_ovhd; /* Cost of the method overhead */
implem deferred FINDX   select_fnidx;
implem deferred RULE    select_rule;
implem deferred PREM    select_prem;

#define INIT_METHOD()  (method_base = method_ovhd = UBINT_C(0))

/* ASSERTIONS */

#define IS_METHOD()  (method && method->Init && method->Select\
		      && method->Retry && method->OkCM)
/* MODIFIERS */

#define MethodInit()   (INIT_METHOD(), method->Init())

#define MethodSelect() method->Select()

#define MethodRetry()  method->Retry()

/* ACCESSORS */

#define MethodOkCM()   method->OkCM()

#define MethodOvhd()   VALUE(method_ovhd)

#define MethodBase()   VALUE(method_base)

#define MethodCost()   (MethodBase() + MethodOvhd())   

#endif


