%{
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "strip.h"

/* GLOBALS */
 
  char *YaccPrompt = "STRIP> "; /* The prompt in interactive mode */
  BOOL  YaccInDef = FALSE;
  PFORM YaccForm = NULL;
  LFORM YaccLForm = LFORM_EMPTY;
  
#define YaccDef(code) (YaccInDef = TRUE,YaccForm = (code),YaccForm)

  void YaccDefine(void);
  void YaccDefForm(void);
  BOOL YaccDelForm(PFORM f);
  void YaccDelLForm(LFORM l);
  void YaccOutPut(void);
  
  int yylex(void);
  int yyerror(char *);
  
  char *OutMode;
  char *OutRdx;
  char *OutExt;
  char OutPath[MAXPATHNAME + 1];
  FILE *StdOut;

#define OUTEXT(ext) (OutExt = (ext))
    
  BOOL (*OutPut)(void);
 
%}

%union {	
  int num;
  char *nom;
  typmet_t met;
  typimp_t imp;
  typtrc_t trc;
  bmask_t bmk;
  PFORM	f;
}

%token <nom> T_VARNAME
%token <nom> T_FILE
%token <num> T_NUMBER
%token T_TOP T_BOT T_AND T_OR T_IMP T_NOT T_EQV T_LPAR T_RPAR T_COMMA T_TSTILE
%token T_EQU T_EOL T_NOP T_HELP T_QUIT
%token T_METHOD T_FLEAF T_RPREC 
%token T_IMPLEM T_ISC T_LRM
%token T_SET T_ITER T_PROMPT T_NORMAL  
%token T_DEFINE T_SEQUENT T_CHECK T_TRACE T_PROVE T_REFUTE T_SOLVE
%token T_OUTW T_OUTA
%token T_PRINT T_VERBOSE T_STAT T_PROOF T_KRIPKE
%token T_MP T_TEX T_XML T_GDL

%nonassoc T_TSTILE
%left T_COMMA
%left T_EQV
%left T_IMP
%left T_AND T_OR 
%left notun

%type <f>   form
%type <imp> implem
%type <met> method
%type <trc> tracemode
%type <bmk> solvemode
%type <bmk> or_refute
%type <bmk> or_prove
 
%start main

%%

main: { } | lcommandline { }
;

lcommandline: commandline { } | lcommandline commandline { }
;

commandline: lcommand T_EOL { 
  if (IsTTY()) printf(YaccPrompt);
  yyerrok;
} 
;

lcommand: command { } | command T_NOP lcommand { } | error lcommand { }
;

command: commands { 
  if (! IsErrOk()) {
    PrintErr(); 
    EraseErr(); 
  } 
}
;

commands: { }
| T_QUIT { YYACCEPT; }
| T_HELP { Help(); }
| T_DEFINE def_seq { }
| search { }
| output { }
;

def_seqf: { } | def_seq { } 
;

def_seq: sequent { YaccDefine(); } 
;

sequent: def_form { } | T_TSTILE def_form { } | def_lform T_TSTILE def_form { }
;
  
def_lform: def_form { } | def_lform T_COMMA def_form { }
;

def_form: form { YaccDefForm(); }
;

form: T_TOP { $$ = YaccDef(NewCst(TopValue)); }
| T_BOT { $$ = YaccDef(NewCst(BotValue)); }
| T_VARNAME { $$ = YaccDef(NewVar((char *) strdup($1))); } 
| form T_EQV form  { $$ = YaccDef(NewEqv($1,$3)); }
| form T_IMP form { $$ = YaccDef(NewImp($1,$3)); }
| form T_OR form  { $$ = YaccDef(NewOr($1,$3)); }
| form T_AND form { $$ = YaccDef(NewAnd($1,$3)); }
| T_NOT form %prec notun { $$ = YaccDef(NewNot($2)); }
| T_LPAR form T_RPAR { $$ = $2; }
;


search: T_CHECK def_seqf { if (Check(Iterator())) PrintResult(); }
| T_TRACE def_seqf {  if (Trace()) PrintResult(); }
| T_PROVE def_seqf { if (Prove()) PrintResult(); }
| T_REFUTE def_seqf { if (Refute()) PrintResult(); }
| T_SOLVE def_seqf { if (Solve()) PrintResult(); }
;

output: outputs { YaccOutPut(); } 
;

outputs: t_print T_MP filename { OutPut = PrintMP; OUTEXT(".mp"); }
| t_print T_XML filename { OutPut = PrintXML; OUTEXT(".xml"); }
| t_print T_GDL filename { OutPut = PrintGDL; OUTEXT(".gdl"); }
| t_print T_TEX filename { OutPut = PrintTEX; OUTEXT(".tex"); }
| t_print T_PROOF filename { OutPut = PrintProof; OUTEXT(".prf"); }
| t_print T_KRIPKE filename { OutPut = PrintKripke; OUTEXT(".cmo"); }
| t_print T_SEQUENT filename { OutPut = PrintResult; OUTEXT(".out"); }
| t_print T_STAT filename { OutPut = PrintStat; OUTEXT(".out"); }
| T_VERBOSE T_STAT filename { OutPut = VerboseStat; OUTEXT(".out"); }
| T_PRINT T_SET filename { OutPut = PrintSet; OUTEXT(".out"); } 
| T_VERBOSE T_SET filename { OutPut = VerboseSet; OUTEXT(".out"); } 
| T_SET filename { OutPut = PrintSet; OUTEXT(".out"); }
| T_SET settings { OutPut = NULL; }
;

t_print: { } | T_PRINT { }
;

filename: { OutRdx = NULL;  } 
| T_OUTW T_FILE { OutMode = "w"; OutRdx = $2; }
| T_OUTA T_FILE { OutMode = "a"; OutRdx = $2; }
;

settings: setting { } | settings T_COMMA setting { }
;

setting:  T_IMPLEM T_EQU implem { SetImplem($3); }
| T_METHOD T_EQU method { SetMethod($3); }
| T_TRACE T_EQU tracemode { SetTraceMode($3); }
| T_SOLVE T_EQU solvemode { SetBuildMask($3); }
| T_ITER T_EQU T_NUMBER { SetIterator((unsigned int) $3); }
;

implem: T_ISC { $$ = TYPIMP_ISC; } | T_LRM { $$ = TYPIMP_LRM; }
;

method : T_FLEAF { $$ = TYPMET_FLEAF; } | T_RPREC { $$ = TYPMET_RPREC; }
;

tracemode: T_NORMAL { $$ = TYPTRC_NORMAL; }
| T_PROMPT { $$ = TYPTRC_PROMPT; }
;

solvemode: T_CHECK { $$ = BMASK_NOTHING; }
| T_PROVE or_refute { $$ = $2; }
| T_REFUTE or_prove { $$ = $2; }
;

or_refute: { $$ = BMASK_PROOF; } | T_OR T_REFUTE { $$ = BMASK_BOTH; }
;

or_prove: { $$ = BMASK_KRIPKE; } | T_OR T_PROVE { $$ = BMASK_BOTH; }
;

%%

void YaccDefForm(void)
{
  YaccLForm = ConsLForm(YaccLForm,YaccForm);
  YaccForm = NULL;
}

BOOL YaccDelForm(PFORM p)
{
  if (p) {
    if (FormIsVar(p)) 
      free((void *) FormValue(p));
    DelForm(p);
  }
  return TRUE;
}

void YaccDelLForm(LFORM l)
{
  MapLForm(l,YaccDelForm);  
  DelLForm(l,FALSE);
}

void YaccDefine(void)
{
  REQUIRE(! LFormIsEmpty(YaccLForm));

  YaccDelLForm(Antecedent()); /* Delete current context */
  YaccDelForm(Conclusion());  /* Delete current conclusion */
  Define(LFormTail(YaccLForm),LFormForm(YaccLForm));
  YaccLForm = LFORM_EMPTY;
  YaccForm = NULL;
  YaccInDef = FALSE;
}


void YaccOutPut(void)
{
  if (OutPut == NULL) return;
  else if (OutRdx == NULL) OutPut();
  else {
    StdOut = stdout;
    GetStripDir(OutPath,MAXPATHNAME);
    StrCatN(OutPath,"/tmp/",MAXPATHNAME);
    StrCatN(OutPath,OutRdx,MAXPATHNAME);
    if ((strchr(OutPath,'.') == NULL) && (OutExt != NULL))
      StrCatN(OutPath,OutExt,MAXPATHNAME);
    stdout = fopen(OutPath,OutMode);
    if (stdout == NULL) {
      stdout = StdOut;
      printf("cannot open file %s !!\n",OutPath); 
    } else {
      OutPut();
      fclose(stdout);
      stdout = StdOut;
      if (IsErrOk())
	printf("output written in %s\n",OutPath);
    }
  }
}

int yyerror(char *s)
{       
  printf("%s: recovering after invalid command\n",s);
  if (YaccInDef) {
    /* deleting erroneously defined formulae */
    YaccDelForm(YaccForm);
    YaccForm = NULL;
    YaccDelLForm(YaccLForm);
    YaccLForm = LFORM_EMPTY;
    YaccInDef = FALSE;
  }
  return 1;
}

int main(int argc,char **argv)
{
  if (IsTTY())
    printf(YaccPrompt);

  MakeSTRIP();
  yyparse();
  KillSTRIP();
  NewLine();
  return 0;
}










