#include "proof.h"

PNODE *ptree_space; /* Memory space to hold proof-tree nodes */
PNODE *ptree_sfree; /* Pointer to the first free node in the space */
PSIZE  ptree_scapa; /* Capacity of the memory space */
PSIZE  ptree_ssize; /* Number of allocated nodes in the memory space */

PNODE *AllocPSpace(ubint_t capa)
{
  PSIZE i;
  PNODE *result;

  REQUIRE(ptree_space == NULL);
  REQUIRE(capa > PSIZE_C(0));

  result = (PNODE *) malloc(sizeof(struct pnode_s) * capa);
  if (result != NULL) {
    ptree_space = result;
    ptree_sfree = result;
    ptree_scapa = capa;
    ptree_ssize = PSIZE_C(0);
    for (i = PSIZE_C(1); i < ptree_scapa; i++)
      {
	PT_SON1(result) = result + 1;
	result++;
      }
    PT_SON1(result) = NULL;
  }
  return result;
}

void FreePSpace()
{
  if (ptree_space != NULL) {
    free(ptree_space);
    ptree_space = NULL;
    ptree_sfree = NULL;
    ptree_scapa = PSIZE_C(0);
    ptree_ssize = PSIZE_C(0);
  }
}

PTREE NewPTree(FINDX fnidx)
     /* This function allocates memory for a new proof-node.
	The new node is initialized with "rule" and "fnidx"
     */
{
  PTREE result = NULL;

  REQUIRE(ptree_space != NULL);
  if (ptree_ssize < ptree_scapa) {
    result = ptree_sfree;
    ptree_sfree = PT_SON1(result);
    ptree_ssize++;
    PT_FINDX(result) = fnidx;
    PT_SON1(result) = PTREE_EMPTY;
    PT_SON2(result) = PTREE_EMPTY;
    PT_FATH(result) = PTREE_EMPTY;
  }
  ENSURE(result != NULL);
  return result;
}

PTREE AddPTree(PTREE tree,PTREE son,PREM prem)
     /* This function adds proof-tree "son" as a 
	son of proof-tree "tree".
	The functions returns the resulting tree
     */
{
  REQUIRE(! PTreeIsEmpty(tree)); /* We need a tree to link to */
  REQUIRE(! PTreeIsEmpty(son));  /* We need a tree to be linked */
  REQUIRE(IS_PREM(prem)); /* The premiss number must be valid */
  /* We cannot link a tree which is already a son of some other tree, 
     and, moreover, we cannot link a tree as the n-th son if there is 
     already exists a n-th son. */
  REQUIRE(PTreeIsEmpty(PT_FATH(son))); 
  REQUIRE(IMP(prem == PREM_FST,PTreeIsEmpty(PT_SON1(tree))));
  REQUIRE(IMP(prem == PREM_SND,PTreeIsEmpty(PT_SON2(tree))));

  PT_FATH(son) = tree;
  if (prem == PREM_FST)
    PT_SON1(tree) = son;
  else
    PT_SON2(tree) = son;
  return tree;
}

PTREE DelPTree(PTREE tree)
     /* This function frees the memory allocated
	for proof-tree "tree".
	The function always returns PTREE_EMPTY 
     */
{
  REQUIRE(ptree_space != NULL);

  if (PTreeIsEmpty(tree))
    return tree;
  /* First remove left son */
  if (PT_SON1(tree) != PTREE_EMPTY) 
    DelPTree(PT_SON1(tree));
  /* Then remove right son */
  if (PT_SON2(tree) != PTREE_EMPTY)
    DelPTree(PT_SON2(tree));

  /* Remove the node itself */

  /* free((void *) tree); */
  ptree_ssize--;
  PT_SON1(tree) = ptree_sfree;
  ptree_sfree = tree;

  return PTREE_EMPTY;
}

ubint_t PTreeSize(PTREE tree)
     /* This function returns the size of the proof-tree */
{
  if (PTreeIsEmpty(tree))
    return PSIZE_C(0);
  else
    return (PTreeSize(PT_SON1(tree)) + 
	    PTreeSize(PT_SON2(tree)) + PSIZE_C(1));
}

#define MAXDEPTH(v1,v2) (((v1) >= (v2)) ? (v1):(v2))

ubint_t PTreeDepth(PTREE tree)
{
  if (PTreeIsEmpty(tree))
    return PSIZE_C(0);
  if (PTreeIsEmpty(PT_SON1(tree)) && PTreeIsEmpty(PT_SON2(tree)))
    return PSIZE_C(0);
  return MAXDEPTH(PTreeDepth(PT_SON1(tree)),
		  PTreeDepth(PT_SON2(tree))) + PSIZE_C(1);
}
