#include "prover.h"

/******************
 * IMPLEMENTATION *
 ******************/

/* PROOF-BUILDING FUNCTIONS */

void ProveStart(void)
{
  CheckStart();

  DelPTree(ProofTree());
  proof_tree = PTREE_EMPTY;
}

void ProveExpandSucc(void) 
{ 
  CheckExpandSucc();

  /* The expansion succeeded. We create a one-node proof-tree
     with the rule and the index that lead to the axiom. */
  PE_PTREE(Top()) = NewPTree(TopIdx());
}

void ProveExpandFail(void)
{
  CheckExpandFail();

  /* The expansion failed. We don't have any proof to build. */
  PE_PTREE(Top()) = PTREE_EMPTY;
}

void ProvePremissFail(void)
{
  CheckPremissFail();

  /* The rule failed. Therefore, if we have built a partial proof due 
     to the success of a previous premiss, we must now destroy it. */   
  DelPTree(TopPTree());
  PE_PTREE(Top()) = PTREE_EMPTY;
}

void ProvePremissSucc(void)
{
 CheckPremissSucc();

  /* The rule succeeded. We have two possible cases:
     1 - we already have a partial proof due to the success of a 
     previous premiss, then, we simply complete it by adding the 
     proof-tree of the reversed rule (second premiss) as a son 
     of the partial proof.
     2 - we don't have any partial proof for the moment. Therefore,
     we add the proof-tree of the reversed rule (first premiss) as
     a son of a newly created one-node tree corresponding to the 
     rule of which the reversed rule is a premiss. For example,
     suppose that we have reversed the premiss A of (A & B). Then, 
     the reversed proof-tree is a proof of A. To obtain the partial
     proof of (A & B), we simply create a one-node proof-tree for 
     (A & B) to which we link, as a son, the proof-tree of A.
  */
 if (PTreeIsEmpty(TopPTree())) {
   PE_PTREE(Top()) = NewPTree(TopIdx());
 }
 PE_PTREE(Top()) = AddPTree(TopPTree(),RevPTree(),RevPrem());
}

void ProveSuccess(void)
{
  CheckSuccess();

  proof_tree = TopPTree(); /* Retrieve proof-tree from the stack */
  proof_typmet = SearchMethod(); /* Keep track of the construction method */
  proof_size = CheckPMem(); /* Compute the proof-size */
  proof_depth = PTreeDepth(ProofTree());  /* Compute proof-depth */

  ENSURE(! PTreeIsEmpty(ProofTree()));
  ENSURE(ProofMethod() == SearchMethod());
  ENSURE(ProofSize() > PSIZE_C(0));
}

void ProveFailure(void)
{
  CheckFailure();

  proof_tree = TopPTree(); /* Should be PTREE_EMPTY */

  ENSURE(PTreeIsEmpty(ProofTree()));
}
