#include "pstack.h"
         
/******************
 * IMPLEMENTATION *
 ******************/

void MakePs(void)
/*  Allocate a new proof-stack with one initial element (cf ResetPs) */
{
  REQUIRE(! IS_PSTACK());

  pstack = (PSTACK) ALLOCMEM(sizeof(struct pstack_s) + 
			     N_NCOUNT * (sizeof(struct pselem_s)));
  if (pstack == NULL)
    FatalError("NewPStack","Cannot allocate proof-stack");

  pstack->item = (PSELEM) (pstack + 1);
  pstack->top = FSIZE_MAX; /* Se top index to -1 */
  ResetPs();

  ENSURE(PsSize() == 1);
}

void ResetPs(void)
{
  REQUIRE(PsSize() <= FSIZE_C(1));

  pstack->top = 0; /* Reset the top index */
  /* We push an initial element on the search-stack for two reasons:
     1 - to avoid testing for the empty stack when executing an operation 
     which is not defined on the empty stack
     2 - to store information (kripke/proof-tree) concerning the 
     initial forest (even when no rule needs to be applied)
  */
  PE_START(Top()) = FINDX_OUT;
  PE_FINDX(Top()) = FINDX_OUT;
  PE_RULE(Top()) = RULE_OUT;
  PE_PREM(Top()) = PREM_OUT;
  PE_PMEM(Top()) = UBINT_C(0);
  PE_KMEM(Top()) = UBINT_C(0);
  PE_PTREE(Top()) = PTREE_EMPTY;
  PE_KTREE(Top()) = KTREE_EMPTY;

  pstack->iden = UBINT_C(1); /* Reset situation identifier generator */
  focus_fnidx = SeqFirstIdx();
}

void KillPs(void)
/* Remove the proof-stack.. */
{
  if (pstack != NULL) FREEMEM(pstack);
}
