#ifndef __PSTACK_H
#define __PSTACK_H

#include "proof.h"
#include "kripke.h"

/************************
 * TYPES AND STRUCTURES *
 ************************/

/* PROOF-STACK ELEMENTS */

typedef struct pselem_s pselem_t, *PSELEM;
struct pselem_s
{
  ubint_t id;  /* Situation identifier */ 
  FINDX start; /* Index of the first developed child */
  FINDX fnidx; /* Index on which the rule has been applied */
  RULE  rule;  /* Logical rule leading to the situation */
  PREM  prem;  /* Index of the rule premiss */ 
  INFO  info;  /* Information associated with the rule */
  ubint_t pmem; /* Memory usage of the local proof-tree */
  ubint_t kmem; /* Memory usage of the local kripke-tree */
  PTREE proof;  /* Local proof-tree of the situation */
  KTREE kripke; /* Local kripke-tree of the situation  */
};

#define PE_SIT(pe)   ((pe)->id)
#define PE_START(pe) ((pe)->start)
#define PE_FINDX(pe) ((pe)->fnidx)
#define PE_RULE(pe)  ((pe)->rule)
#define PE_PREM(pe)  ((pe)->prem)
#define PE_INFO(pe)  ((pe)->info)
#define PE_PTREE(pe) ((pe)->proof)
#define PE_KTREE(pe) ((pe)->kripke)
#define PE_PMEM(pe)  ((pe)->pmem)
#define PE_KMEM(pe)  ((pe)->kmem)

/* PROOF-STACK STRUCTURE */

typedef struct pstack_s pstack_t, *PSTACK;
struct pstack_s 
{
  ubint_t iden; /* Last generated situation identifier */
  FSIZE top;
  pselem_t *item; /* Array of proof-stack elements */
};


/*************
 * INTERFACE *
 *************/

implem FINDX  focus_fnidx; /* A valid index in the current forest */
implem PSTACK pstack; /* Current proof-search stack */

/* ASSERTIONS */

#define IS_PSTACK() (pstack != NULL)

/* CONSTRUCTORS  - DESTRUCTORS */

public void MakePs(void);

public void KillPs(void);

/* ACCESSORS */

#define PsCapa()   N_NCOUNT
#define PsSize()   GETSAFE(IS_PSTACK(),pstack->top + FSIZE_C(1))
#define PsSit()    GETSAFE(IS_PSTACK(),pstack->iden)

#define Top()      GETSAFE(IS_PSTACK(),(pstack->item + pstack->top))
#define TopSit()   PE_SIT(Top())
#define TopIdx()   PE_FINDX(Top())
#define TopRule()  PE_RULE(Top())
#define TopPrem()  PE_PREM(Top())
#define TopPMem()  PE_PMEM(Top())
#define TopKMem()  PE_KMEM(Top())
#define TopPTree() PE_PTREE(Top())
#define TopKTree() PE_KTREE(Top())

#define RevSit()   PE_SIT(Top() + FSIZE_C(1))
#define RevIdx()   PE_FINDX(Top() + FSIZE_C(1))
#define RevRule()  PE_RULE(Top() + FSIZE_C(1))
#define RevPrem()  PE_PREM(Top() + FSIZE_C(1))
#define RevPMem()  PE_PMEM(Top() + FSIZE_C(1))
#define RevKMem()  PE_KMEM(Top() + FSIZE_C(1))
#define RevPTree() PE_PTREE(Top() + FSIZE_C(1))
#define RevKTree() PE_KTREE(Top() + FSIZE_C(1))


/* MODIFIERS */

public void ResetPs(void);

#define PushAndApply(rule,prem,fnidx) {\
  REQUIRE(PsSize() < PsCapa());\
  \
  pstack->top++; \
  PE_SIT(Top()) = ++pstack->iden;\
  PE_START(Top()) = FINDX_OUT;\
  PE_FINDX(Top()) = (fnidx);\
  PE_RULE(Top()) = (rule);\
  PE_PREM(Top()) = (prem); \
  PE_PMEM(Top()) = UBINT_C(0);\
  PE_KMEM(Top()) = UBINT_C(0);\
  PE_PTREE(Top()) = PTREE_EMPTY;\
  PE_KTREE(Top()) = KTREE_EMPTY;\
  focus_fnidx = RuleApply((rule),(prem),(fnidx),PE_INFO(Top()));\
}

#define ReverseAndPop() {\
  REQUIRE(PsSize() > FSIZE_C(1)); \
  \
  RuleReverse(TopRule(),TopPrem(),TopIdx(),PE_INFO(Top())); \
  pstack->top--; \
}

#endif




