#include "rules.h"

/******************
 * IMPLEMENTATION *
 ******************/

/*******************************************
 * The following rules are intended to be  *
 * applied on root leaves (one node trees) *
 *******************************************/

/* Conclusion rules */

FINDX Apply_r_and_1(FINDX conclu,INFO infos) 
{ 
  N_CONCLU = N_LSF(conclu);
  return N_LSF(conclu); 
}

void Rever_r_and_1(FINDX node,INFO infos)
{ 
  N_CONCLU = node; 
}

FINDX Apply_r_and_2(FINDX conclu,INFO infos)
{ 
  N_CONCLU = N_RSF(conclu);
  return N_RSF(conclu); 
}

void Rever_r_and_2(FINDX node,INFO infos)
{ 
  N_CONCLU = node; 
}

FINDX Apply_r_or_1(FINDX conclu,INFO infos)
{ 
  N_CONCLU = N_LSF(conclu);
  return N_LSF(conclu); 
}

void Rever_r_or_1(FINDX node,INFO infos)
{ 
  N_CONCLU = node; 
}

FINDX Apply_r_or_2(FINDX conclu,INFO infos) 
{ 
  N_CONCLU = N_RSF(conclu);
  return N_RSF(conclu); 
}

void Rever_r_or_2(FINDX node,INFO infos)
{ 
  N_CONCLU = node; 
}

FINDX Apply_r_imp(FINDX conclu,INFO infos) 
{ 
  ForestNewRootL(N_LSF(conclu));
  EnterContext(N_LSF(conclu));
  N_CONCLU = N_RSF(conclu);
  return N_LSF(conclu);
}

void Rever_r_imp(FINDX node,INFO infos)
{
  LeaveContext(N_LSF(node));
  ForestRemoveL(N_LSF(node));
  N_CONCLU = node;
}

/* Root Rules */

FINDX Apply_l_and(FINDX leaf,INFO infos) 
{ 
  ForestSplitLH(leaf,N_LSF(leaf),N_RSF(leaf));
  EnterContext(N_LSF(leaf));
  EnterContext(N_RSF(leaf));
  return N_LSF(leaf);
}
void Rever_l_and(FINDX node,INFO infos)
{
  ForestUnsplitLH(node,N_LSF(node),N_RSF(node));
  LeaveContext(N_LSF(node));
  LeaveContext(N_RSF(node));
}

FINDX Apply_l_or_1(FINDX leaf,INFO infos) 
{ 
  ForestReplaceL(leaf,N_LSF(leaf));
  EnterContext(N_LSF(leaf));
  return N_LSF(leaf); 
}

void Rever_l_or_1(FINDX node,INFO infos)
{ 
  ForestReplaceL(N_LSF(node),node);
  LeaveContext(N_LSF(node));
}

FINDX Apply_l_or_2(FINDX leaf,INFO infos) 
{ 
  ForestReplaceL(leaf,N_RSF(leaf)); 
  EnterContext(N_RSF(leaf));
  return N_RSF(leaf); 
}

void Rever_l_or_2(FINDX node,INFO infos)
{ 
  ForestReplaceL(N_RSF(node),node); 
  LeaveContext(N_RSF(node));
}

FINDX Apply_l_imp(FINDX leaf,INFO infos) 
/* This rule doesn't really exist in SLJ */
{ 
  ForestSplitLV(leaf,N_RSF(leaf),N_LSF(leaf));
  return N_LSF(leaf);
}

void Rever_l_imp(FINDX node,INFO infos)
{ 
  ForestUnsplitLV(node,N_RSF(node),N_LSF(node)); 
}

/**********************************************
 * The following rules are intended to be     *
 * applied on strict leaves (having a father) *
 **********************************************/

/* Non-starred leaf rules */

FINDX Apply_or_imp(FINDX leaf,INFO infos) 
{ 
  ForestSplitLH(leaf,N_LSF(leaf),N_RSF(leaf));
  return N_LSF(leaf); 
}

void Rever_or_imp(FINDX node,INFO infos)
{ 
  ForestUnsplitLH(node,N_LSF(node),N_RSF(node)); 
}

FINDX Apply_and_imp(FINDX leaf,INFO infos) 
{ 
  ForestSplitLV(leaf,N_RSF(leaf),N_LSF(leaf));
  return N_LSF(leaf); 
}

void Rever_and_imp(FINDX node,INFO infos)
{ 
  ForestUnsplitLV(node,N_RSF(node),N_LSF(node)); 
}

FINDX Apply_imp_imp_1(FINDX leaf,INFO infos) /* non-invertible premiss */
{
  ForestReplaceL(leaf,N_RSF(leaf));
  ForestNewRootL(N_LSF(leaf));
  EnterContext(N_LSF(leaf));
  INFO_TAGGED(infos) = N_TAGGED;
  INFO_CONCLU(infos) = N_CONCLU;
  N_TAGGED = N_RSF(leaf);
  N_CONCLU = FINDX_OUT;
  return N_RSF(leaf);
} 

void Rever_imp_imp_1(FINDX node,INFO infos)
{
  N_CONCLU = INFO_CONCLU(infos);
  N_TAGGED = INFO_TAGGED(infos);
  LeaveContext(N_LSF(node));
  ForestRemoveL(N_LSF(node));
  ForestReplaceL(N_RSF(node),node);
}

FINDX ApplyTreeCut(FINDX leaf,INFO infos) /* need information */
{
  FINDX father = N_FA(leaf);

  ForestDetachST(leaf,&INFO_FSTSON(infos),\
		 &INFO_LMLEAF(infos),&INFO_RMLEAF(infos));

  if ((INFO_LMLEAF(infos) > N_TAGGED) || (N_TAGGED > INFO_RMLEAF(infos)))
    /* Say that we haven't done anything on the tag */
    INFO_TAGGED(infos) = FINDX_OUT;
  else {
    /* Since the tagged node will be cut, it becomes the conclusion */
    INFO_TAGGED(infos) = N_TAGGED;
    N_CONCLU = N_TAGGED;
    N_TAGGED = FINDX_OUT;
  }

  /* The father is now a leaf */
  if (IdxIsRoot(father))
    /* The father appears now also as a root */
    EnterContext(father);
  return father;
}

void ReverseTreeCut(FINDX node,INFO infos)
{
  FINDX father = N_FA(node);

  if (IdxIsRoot(father))
    LeaveContext(father);

  if (INFO_TAGGED(infos) != FINDX_OUT) {
    N_TAGGED = INFO_TAGGED(infos);
    N_CONCLU = FINDX_OUT;
  }

  ForestAttachST(node,INFO_FSTSON(infos),
		 INFO_LMLEAF(infos),INFO_RMLEAF(infos));
}

#define Apply_imp_imp_2 ApplyTreeCut
#define Apply_atom_imp ApplyTreeCut

#define Rever_imp_imp_2 ReverseTreeCut
#define Rever_atom_imp  ReverseTreeCut

/* Starred leaves rules */

FINDX Apply_or_s_imp_1(FINDX leaf,INFO infos)
{
  ForestSplitLH(leaf,N_LSF(leaf),N_RSF(leaf));
  N_TAGGED = N_LSF(leaf);
  return N_LSF(leaf);
}
void Rever_or_s_imp_1(FINDX node,INFO infos)
{
  ForestUnsplitLH(node,N_LSF(node),N_RSF(node));
  N_TAGGED = node;
}

FINDX Apply_or_s_imp_2(FINDX leaf,INFO infos)
{
  ForestSplitLH(leaf,N_LSF(leaf),N_RSF(leaf));
  N_TAGGED = N_RSF(leaf);
  return N_LSF(leaf);
}

void Rever_or_s_imp_2(FINDX node,INFO infos)
{
  ForestUnsplitLH(node,N_LSF(node),N_RSF(node));
  N_TAGGED = node;
}

FINDX Apply_and_s_imp_1(FINDX leaf,INFO infos)
{
  ForestReplaceL(leaf,N_LSF(leaf));
  N_TAGGED = N_LSF(leaf);
  return N_LSF(leaf);
}

void Rever_and_s_imp_1(FINDX node,INFO infos)
{
  ForestReplaceL(N_LSF(node),node);
  N_TAGGED = node;
}

FINDX Apply_and_s_imp_2(FINDX leaf,INFO infos)
{
  ForestReplaceL(leaf,N_RSF(leaf));
  N_TAGGED = N_RSF(leaf);
  return N_RSF(leaf);
}

void Rever_and_s_imp_2(FINDX node,INFO infos)
{
  ForestReplaceL(N_RSF(node),node);
  N_TAGGED = node;
}

FINDX Apply_imp_s_imp(FINDX leaf,INFO infos)
{
  ForestReplaceL(leaf,N_RSF(leaf));
  ForestNewRootL(N_LSF(leaf));
  EnterContext(N_LSF(leaf));
  N_TAGGED = N_RSF(leaf);
  return N_RSF(leaf);
}

void Rever_imp_s_imp(FINDX node,INFO infos)
{
  N_TAGGED = node;
  LeaveContext(N_LSF(node));
  ForestRemoveL(N_LSF(node));
  ForestReplaceL(N_RSF(node),node);
}

/* CONSTRUCTORS - DESTRUCTORS */

#define APP(r) (Apply_ ## r)
#define REV(r) (Rever_ ## r)
#define APP1(r) (Apply_ ## r ## _1)
#define APP2(r) (Apply_ ## r ## _2)
#define REV1(r) (Rever_ ## r ## _1)
#define REV2(r) (Rever_ ## r ## _2)

rule_t trule[RULE_MAX + 1] = {
  /* Conclusion rules */
  { "R&", CONJUNCT, RPREC_C(0), 
    { { TRUE, APP1(r_and), REV1(r_and) },
      { TRUE, APP2(r_and), REV2(r_and) }
    }
  },
  { "R|", DISJUNCT, RPREC_C(0), 
    { { FALSE, APP1(r_or), REV1(r_or) },
      { FALSE, APP2(r_or), REV2(r_or) }
    }
  },
  { "R->", SINGLE, RPREC_C(0), 
    { { TRUE, APP(r_imp), REV(r_imp) }
    }
  },
  /* Root rules */
  { "L&", SINGLE, RPREC_C(0),
    { { TRUE, APP(l_and), REV(l_and) }
    }
  },
  { "L|", CONJUNCT, RPREC_C(0),
    { { TRUE, APP1(l_or), REV1(l_or) },
      { TRUE, APP2(l_or), REV2(l_or) }
    }
  },
  { "L->", SINGLE, RPREC_C(0),
    { { TRUE, APP(l_imp), REV(l_imp) }
    }
  },
  /* Non-starred leaf rules */
  { "&->", SINGLE, RPREC_C(0),
    { { TRUE, APP(and_imp), REV(and_imp) }
    }
  },
  { "|->", SINGLE, RPREC_C(0),
    { { TRUE, APP(or_imp), REV(or_imp) }
    }
  },
  { "->->", CONJUNCT, RPREC_C(0),
    { { FALSE, APP1(imp_imp), REV1(imp_imp) },
      { TRUE,  APP2(imp_imp), REV2(imp_imp) }
    }
  },
  { "X->", SINGLE, RPREC_C(0),
    { { TRUE, APP(atom_imp), REV(atom_imp) }
    }
  },
  /*  Starred leaf rules */
  { "&*->", CONJUNCT, RPREC_C(0),
    { { TRUE, APP1(and_s_imp), REV1(and_s_imp) },
      { TRUE,  APP2(and_s_imp), REV2(and_s_imp) }
    }
  },
  { "|*->", DISJUNCT, RPREC_C(0),
    { { FALSE, APP1(or_s_imp), REV1(or_s_imp) },
      { FALSE, APP2(or_s_imp), REV2(or_s_imp) }
    }
  },
  { "->*->", SINGLE, RPREC_C(0),
    { { TRUE, APP(imp_s_imp), REV(imp_s_imp) }
    }
  }
};  /* Array of rules */

/* Optimisation: access to rules by fnidx-type */
RULE conc_rule[FTYPE_MAX + 1];
RULE root_rule[FTYPE_MAX + 1];
RULE leaf_rule[FTYPE_MAX + 1];
RULE star_rule[FTYPE_MAX + 1];

void MakeRules(void)
{
  /* Optimisation: classify rules by fnidx-type */
  conc_rule[FTYPE_TO_CINDX(FTYPE_CST)] = RULE_OUT;
  conc_rule[FTYPE_TO_CINDX(FTYPE_VAR)] = RULE_OUT;
  conc_rule[FTYPE_TO_CINDX(FTYPE_AND)] = r_and;
  conc_rule[FTYPE_TO_CINDX(FTYPE_IMP)] = r_imp;
  conc_rule[FTYPE_TO_CINDX(FTYPE_OR)] = r_or;

  root_rule[FTYPE_TO_CINDX(FTYPE_CST)] = RULE_OUT;
  root_rule[FTYPE_TO_CINDX(FTYPE_VAR)] = RULE_OUT;
  root_rule[FTYPE_TO_CINDX(FTYPE_AND)] = l_and;
  root_rule[FTYPE_TO_CINDX(FTYPE_IMP)] = l_imp;
  root_rule[FTYPE_TO_CINDX(FTYPE_OR)] = l_or;

  star_rule[FTYPE_TO_CINDX(FTYPE_CST)] = RULE_OUT;
  star_rule[FTYPE_TO_CINDX(FTYPE_VAR)] = RULE_OUT;
  star_rule[FTYPE_TO_CINDX(FTYPE_AND)] = and_s_imp;
  star_rule[FTYPE_TO_CINDX(FTYPE_OR)] = or_s_imp;
  star_rule[FTYPE_TO_CINDX(FTYPE_IMP)] = imp_s_imp;

  leaf_rule[FTYPE_TO_CINDX(FTYPE_CST)] = RULE_OUT;
  leaf_rule[FTYPE_TO_CINDX(FTYPE_VAR)] = RULE_OUT;
  leaf_rule[FTYPE_TO_CINDX(FTYPE_AND)] = and_imp;
  leaf_rule[FTYPE_TO_CINDX(FTYPE_OR)] = or_imp;
  leaf_rule[FTYPE_TO_CINDX(FTYPE_IMP)] = imp_imp;
}

/* ACCESSORS */

RULE FindRule(FINDX fnidx)
{
  static RULE result;
  FTYPE ty;

  REQUIRE(IS_FNODE(fnidx));

  ty = FTYPE_TO_CINDX(N_TY(fnidx));
  if (IdxIsConclu(fnidx))
    result = conc_rule[ty];
  else if (IdxIsRoot(fnidx))
    result = root_rule[ty];
  else if (IdxIsTagged(fnidx))
    result = star_rule[ty];
  else if (ContextHasIdx(fnidx))
    result = atom_imp;  /* We preferentially apply X, X->Y */
  else
    result = leaf_rule[ty];

  ENSURE(IMP(result != RULE_OUT,IS_RULE(result)));
  return result;
}

