#include "strip.h"
#include "help.h"

size_t GetStripDir(char *dest,size_t n)
{
  char *envvar;
  size_t result;

  envvar = getenv("STRIPDIR");
  if (envvar == NULL) 
    result = StrCpyN(dest,".",n);
  else { 
    StrCpyN(dest,envvar,n); /* Copy the environment variable */
    result = StrFmtPath(dest); /* Erase multiple slashes */
    /* Erase last slash, if any */
    if ((result > 0) && (dest[result - 1] == '/')) {
      result--;
      dest[result] = '\0';
    }
  }
  return result;
}

BOOL Help(void)
{
  static char line[MAXPATHNAME + 1];
  FILE *help;
  size_t i;

  i = GetStripDir(line,MAXPATHNAME);
  StrCatN(line + i,"/HELP",MAXPATHNAME - i);

  help = fopen(line,"r");
  if (help == NULL) {
    printf("%s", strip_help);
    RetOk();
    //    RetErr(ERR_NOHELP);
  }
  
  strcpy(line,"");
  while (! feof(help))
    {
      printf("%s",line);
      fgets(line,256,help);
    }
  fclose(help);
  RetOk();
}

void PrintErr(void)
{
  switch (LastErr()) {
  case ERR_OK:
    break;
  case ERR_NOFOREST:
    puts("no sequent has been defined !!"); break;
  case ERR_NORESULT:
    puts("the sequent is not solved yet !!"); break;
  case ERR_NOHELP:   
    puts("no help file available !!"); break;
  case ERR_NOPROOF:  
    puts("the sequent is unprovable !!"); break;
  case ERR_NOMODEL:  
    puts("the sequent is provable !!"); break;
  case ERR_NOMEMORY: 
    puts("not enough memory !!"); break;
  case ERR_BADMETCM: 
    puts("bad method for counter-model construction !!"); break;
  case ERR_UNPROVED:
    puts("the sequent unproved !!"); break;
  case ERR_UNREFUTED:
    puts("the sequent is unrefuted !!"); break;
  case ERR_IO:
    puts("cannot open file !!"); break;    
  }
}

BOOL PrintSet(void)
{
  printf("implem = %s, method = %s, " 
	 "solve = %s, trace = %s, "
	 "iter = %" PRIuBI "\n",
	 ImplemName(Implem()),
	 MethodName(Method()),
	 BuildMaskStr(),
	 TraceModeStr(),
	 Iterator());
  RetOk();
}

BOOL VerboseSet(void)
{
  printf("forest implementation is %s\n",ImplemName(Implem()));
  printf("proof-search method is %s\n",MethodName(Method()));
  printf("solve mode is %s\n",BuildMaskStr());
  printf("trace mode is %s\n",TraceModeStr());
  printf("check iterator %" PRIuBI "\n",Iterator());
  RetOk();
}

BOOL PrintResult(void)
{
  if (! IS_FOREST())
    RetErr(ERR_NOFOREST);

  if (! IS_TYPRES(SearchResult()))
    printf("unknown: ");
  else
    printf("%s: ",SearchResultStr());
  PrintForest(FALSE);
  NewLine();
  RetOk();
}

BOOL PrintStat(void)
{
 if (! IS_FOREST())
    RetErr(ERR_NOFOREST);

  if (! IS_TYPRES(SearchResult()))
    RetErr(ERR_NORESULT);

  printf("Results:\n");
  printf("SR = %-11s   SI = %-11s   SM = %-11s\n",
	 SearchResultStr(),
	 ImplemName(SearchImplem()),
	 MethodName(SearchMethod()));

  printf("SS = %-11" PRIuBI "   SD = %-11" PRIuBI 
	 "   ST = %-11" PRIMILSEC "   SL = %-9" "u" "\n",
	 CheckSize(),CheckDepth(),SearchTime(),SearchLoops());
    
  printf("FB = %-11" PRIuBI "   MB = %-11" PRIuBI "\n",
	 ForestBase(),MethodBase());
  printf("FO = %-11" PRIuBI "   MO = %-11" PRIuBI "\n",
	 ForestOvhd(),MethodOvhd());
  printf("FC = %-11" PRIuBI "   MC = %-11" PRIuBI "   SC = %-11" PRIuBI "\n",
	 ForestCost(),MethodCost(),CheckCost());
  printf("FB/FC = %6.2f %%   MB/MC = %6.2f %%   FC/SC = %6.2f %%\n",
	 PercFBase(), PercMBase(), PercFCost());
  printf("FO/FC = %6.2f %%   MO/MC = %6.2f %%   MC/SC = %6.2f %%\n",
	 PercFOvhd(), PercMOvhd(), PercMCost());

  switch (SearchResult())
    {
    case SUCCESS:
      printf("PU = %-11" PRIuBI "   ",CheckPMem());
      if (PTreeIsEmpty(ProofTree()))
	printf("PS = %-11s   PD = %-11s   PM = %-11s\n",
	       "unknown","unknown","unknown");
      else
	printf("PS = %-11" PRIuBI "   PD = %-11" PRIuBI "   PM = %-11s\n",
	       ProofSize(),ProofDepth(),MethodName(ProofMethod()));
      break;
    case FAILURE:
      printf("KU = %-11" PRIuBI "   ",CheckKMem());
      if (KTreeIsEmpty(KripkeTree()))
	printf("KS = %-11s   KD = %-11s   KM = %-11s\n",
	       "unknown","unknown","unknown");
      else
	printf("KS = %-11" PRIuBI "   KD = %-11" PRIuBI "   KM = %-11s\n",
	       KripkeSize(),KripkeDepth(),MethodName(KripkeMethod()));
    }
  RetOk();
}

BOOL VerboseStat(void)
{
  if (! IS_FOREST())
    RetErr(ERR_NOFOREST);

  if (! IS_TYPRES(SearchResult()))
    RetErr(ERR_NORESULT);

  printf("Results:\n");

  printf(" search result = %s\n",SearchResultStr());
  printf(" search implem = %s\n",ImplemName(SearchImplem()));
  printf(" search method = %s\n",MethodName(SearchMethod()));
  printf(" search space = %" PRIuBI " rules applied\n",CheckSize());
  printf(" search depth = %" PRIuBI "\n",CheckDepth());
  printf(" search loops = %" "u" "\n",SearchLoops());
  printf(" search time  = %" PRIMILSEC "e-3 seconds\n",SearchTime());
  printf(" forest base = %" PRIuBI "\n",ForestBase());
  printf(" forest ovhd = %" PRIuBI "\n",ForestOvhd());
  printf(" forest cost (base + ovhd) = %" PRIuBI "\n",ForestCost());
  printf(" method base = %" PRIuBI "\n",MethodBase());
  printf(" method ovhd = %" PRIuBI "\n",MethodOvhd());
  printf(" method cost (base + ovhd) = %" PRIuBI "\n",MethodCost());
  printf(" search cost (forest + method) = %" PRIuBI "\n",CheckCost()); 
  printf(" forest base / forest_cost = %6.2f%%\n",PercFBase());
  printf(" forest ovhd / forest cost = %6.2f%%\n",PercFOvhd());
  printf(" forest cost / search cost = %6.2f%%\n",PercFCost()); 
  printf(" method base / method cost = %6.2f%%\n",PercMBase());
  printf(" method ovhd / method cost = %6.2f%%\n",PercMOvhd()); 
  printf(" method cost / search cost = %6.2f%%\n",PercMCost()); 

  switch (SearchResult())
    {
    case SUCCESS:
      printf(" proof usage = %" PRIuBI "\n",CheckPMem()); 
      if (PTreeIsEmpty(ProofTree())) {
	printf(" proof method = unknown\n");
	printf(" proof size   = unknown\n");
	printf(" proof depth  = unknown\n");
      } else {
	printf(" proof method = %s\n",MethodName(ProofMethod()));
	printf(" proof size   = %" PRIuBI "\n",ProofSize());
	printf(" proof depth  = %" PRIuBI "\n",ProofDepth());
      }
      break;
    case FAILURE:
      printf(" kripke usage = %" PRIuBI "\n",CheckKMem()); 
      if (KTreeIsEmpty(KripkeTree())) {
	printf(" kripke method = unknown\n");
	printf(" kripke size   = unknown\n");
	printf(" kripke depth  = unknown\n");
      }
      else {
	printf(" kripke method = %s\n",MethodName(KripkeMethod()));
	printf(" kripke size   = %" PRIuBI "\n",KripkeSize());
	printf(" kripke depth  = %" PRIuBI "\n",KripkeDepth());
      }
    }
  RetOk();
}

BOOL PrintKripke(void)
{
  if (! IS_FOREST())
    RetErr(ERR_NOFOREST); 

  if (! IS_TYPRES(SearchResult()))
    RetErr(ERR_NORESULT);

  if (SearchResult() == SUCCESS)
    RetErr(ERR_NOMODEL);

  if (KTreeIsEmpty(KripkeTree()))
    RetErr(ERR_UNREFUTED);

  printf("Counter-model for ");
  PrintForest(FALSE);
  printf("\n\n");
  PrintKTree(KripkeTree());
  printf("\n");
  RetOk();
}

/*********************************************************************/

void PrintProofLine(char *text,unsigned int step,FINDX hilite)
{
  PrintSpace(step);
  printf("%-5s ",text);
  PrintSequent(hilite);
  printf("\n");
}

BOOL PrintProof(void)
{
  BOOL dupli;
  FINDX fnidx;
  PTREE ptree;
  static unsigned int step = 0U;

  if (! IS_FOREST())
    RetErr(ERR_NOFOREST);

  if (! IS_TYPRES(SearchResult()))
    RetErr(ERR_NORESULT);

  if (SearchResult() == FAILURE)
    RetErr(ERR_NOPROOF);
  
  if (PTreeIsEmpty(ProofTree()))
    RetErr(ERR_UNPROVED);
  
  printf("SLJ-Proof for ");
  PrintForest(FALSE);
  printf("\n\n");

  // START:
  ResetPs();
  ptree = ProofTree();
  
 TESTAX:
  if (PT_SON1(ptree) != NULL) goto SELECT;
  else goto EXPSUCC;
  
 SELECT:
  ptree = PT_SON1(ptree);
  select_fnidx = PT_FINDX(ptree);
  select_rule = FindRule(select_fnidx);
  select_prem = PREM_FST;
  if (! RuleIsSing(select_rule)) step++;
  
 EXPAND:
  PushAndApply(select_rule,select_prem,select_fnidx);
  N_PASS(select_fnidx)++; /* Add one visit to the selected index */
  goto TESTAX;
  
 EXPSUCC:
  if (! ContextHasBot()) {
    dupli = TRUE;
    fnidx = N_FIRST_LEAF;
    while (dupli && (fnidx != FINDX_OUT))
      { 
	dupli = ! (IdxIsRoot(fnidx) && (N_CL(fnidx) == N_CL(N_GOAL)));
	fnidx = N_RL(fnidx);
      }
    if (dupli) { 
      NC_DUP(N_GOAL) = UBINT_C(1); 
    }
  }
  PrintProofLine("axiom",step,FINDX_OUT);

 REVSUCC:
  if (PsSize() <= 1) goto SUCCESS;

  ReverseAndPop();
  ptree = PT_FATH(ptree);

  if (RuleIsConj(RevRule()) && (N_PASS(RevIdx()) == UINT8_C(1))) { 
    ptree = PT_SON2(ptree);
    select_fnidx = RevIdx(); /* Re-expand the same index */
    select_rule = RevRule(); /* Using the same rule */
    select_prem = PREM_SND; /* With the other premiss */
    NewLine();
    goto EXPAND;
  } 
  else {
    N_PASS(RevIdx()) = UINT8_C(0); /* Reset index pass-count */
    if (! RuleIsSing(RevRule())) { 
      NewLine();
      step--;
    }
    if (IdxIsRoot(RevIdx()) && (NC_DUP(RevIdx()) > UBINT_C(0))) {
      PrintProofLine(RuleName(RevRule()),step,RevIdx());
      NC_DUP(RevIdx()) = UBINT_C(0);
      PrintProofLine("Dup",step,RevIdx());
    }
    else PrintProofLine(RuleName(RevRule()),step,RevIdx());
    goto REVSUCC;
  }
  
 SUCCESS:
  printf("\n");
  RetOk();
}

/**************
 * CONVERTERS *
 **************/

BOOL PrintTEX(void)
{
  puts("not implemented yet !!");
  RetOk();
}

BOOL PrintMP(void)
{
  switch (SearchResult()) {
  case SUCCESS:
    puts("not implemented yet !!"); 
    RetOk();
  case FAILURE:
    if (KTreeIsEmpty(KripkeTree()))
      RetErr(ERR_UNREFUTED);
  }
  
  switch (SearchResult()) {
  case SUCCESS: 
    puts("not implemented yet !!"); 
    break;
  case FAILURE:
    KTreeToMP(KripkeTree()); 
    break; /* Output metapost on stdout */
  }

  RetOk();
}

BOOL PrintGDL(void)
{
  switch (SearchResult()) {
  case SUCCESS:
    puts("not implemented yet !!"); 
    RetOk();
  case FAILURE:
    if (KTreeIsEmpty(KripkeTree()))
      RetErr(ERR_UNREFUTED);
  }
  
  switch (SearchResult()) {
  case SUCCESS: 
    puts("not implemented yet !!"); 
    break;
  case FAILURE:
    KTreeToGDL(KripkeTree()); 
    break; 
  }

  RetOk();
}

BOOL PrintXML(void)
{
  switch (SearchResult()) {
  case SUCCESS:
    puts("not implemented yet !!"); 
    RetOk();
  case FAILURE:
    if (KTreeIsEmpty(KripkeTree()))
      RetErr(ERR_UNREFUTED);
  }
  
  switch (SearchResult()) {
  case SUCCESS: 
    puts("not implemented yet !!"); 
    break;
  case FAILURE:
    KTreeToXML(KripkeTree()); 
    break; 
  }

  RetOk();
}
